#!/usr/bin/perl
# -*- perl -*-
eval 'exec /usr/bin/perl -S $0 ${1+"$@"}'
    if 0;

## use IPC::Run qw/start run finish/;
use strict;
use Carp;
use Memoize;
use UNIVERSAL qw(isa can);

use TAG;
use TAG::LP;

use AppConfig qw/:argcount :expand/;

use IPC::Run qw/start/;

use POSIX qw(locale_h);
use locale;
setlocale(LC_ALL,"fr");

use Data::Dumper;
use IO::Handle;

use File::Temp qw/tempfile/;

#my $easydir = "/home/graves/clerger/.build/easy";

BEGIN {
  push(@INC,"/home/graves/clerger/Grammars/frmg");
}

use MyRegExp;

my $package= "/home/graves/clerger/Grammars/frmg";
my $lexdir = "/home/graves/clerger/.build/lexiques/easy";

my $config = AppConfig->new(
			    "dicopath|p=f" => {
					       DEFAULT => "$lexdir"
					      },
			    "dico|d=s" => {
					   DEFAULT => 'dico.xlfg'
					  },
			    "macro|m=f" => {DEFAULT => "$lexdir/dico.xlfg.templates"},
			    "lexed|l=f" => {DEFAULT => "/usr/local/bin/lexed"},
			    "map=f" => {DEFAULT => "$package/frgram.map"},
			    "cache|c=f",
			    "expand!" => {DEFAULT => 1},
			    "locutions=f" => {DEFAULT => "$lexdir/dico.xlfg.cmpnd"},
			    "prefixes=f" => {DEFAULT => "$lexdir/dico.sxspell.pref"},
			    "vcomp_expand!" => {DEFAULT => 1},
			    "tmpdir=f" => {DEFAULT => "/tmp"}
			   );

my $conffile;

foreach ("./.frmg.conf",
         "/etc/frmg.conf" ) {
  $conffile=$_, last if (POSIX::access($_,&POSIX::R_OK));
}

# read configuration file
if ($conffile) {
  $config->file("$conffile")
    || die "can't open or process configuration file $conffile";
}

$config->args();

my $dicopath = $config->dicopath();
my $dico = $config->dico();
my $macrofile = $config->macro();
my $expand = $config->expand();
my %unknown = ();

my %cachelexed = ();

my $map = $TAG::LP::map = {};

parse_map_file($map,$config->map());

TAG::LP::header_reader("$package/".$map->{'lp'}{'header'}) if (defined $map->{'lp'}{'header'});

memoize 'macro_expand';
memoize 'entry_parse';

my @lexed = ($config->lexed,'-d',"$dicopath",'-p',"$dico",'consult');

## print "CMD @lexed\n";

my %locutions = ();

if (-r $config->locutions()) {
  open LOC,"<",$config->locutions() 
    ||  die "can't open locution file: $!";
  while (<LOC>) {
    chomp;
    my @seq = split(/\s+/,$_);
    while (@seq) {
      my $prefix = join(' ',@seq);
      $prefix =~ s/' /'/og;
      $locutions{$prefix} ||= 1;
      pop @seq;
    }
  }
  close LOC;
}

my %prefixes = ();

if (-r $config->prefixes()) {
  open PREF,"<",$config->prefixes() 
    ||  die "can't open prefixes file: $!";
  while (<PREF>) {
    next if m{^//};
    next if /^\s*$/;
    chomp;
    my ($pref) = split(/\s+/,$_);
    $prefixes{$pref} = 1;
  }
  close PREF;
}

my $prefpat = join('|',sort {length($b) <=> length($a)} keys %prefixes);
$prefpat = qr/$prefpat/;

my %agglutinates = ( "au" => ":prep le__det",
		     "aux" => ":prep les__det",
		     "auquel" => ":prep lequel",
		     "auxquels" => ":prep lesquels",
		     "auxquelles" => ":prep lesquelles",
		     "du" => "de:prep le__det",
		     "duquel" => "de:prep lequel",
		     "des" => "de:prep les__det",
		     "desquels" => "de:prep lesquels",
		     "desquelles" => "de:prep lesquelles",
		     "s" => "en:prep les__det",
		     "audit" => ":prep ledit",
		     "auxdits" => ":prep lesdits",
		     "auxdites" => ":prep lesdites",
		     "dudit" => "de:prep ledit",
		     "desdits" => "de:prep lesdits",
		     "desdites" => "de:prep lesdites",
		    );
my $agglpat = join('|',keys %agglutinates);
$agglpat = qr/$agglpat/i;

my %macro = ();

my %allowed_cat = ( 'n' => 1,
		    'v' => 1,
		    'cl' => 1,
		    'd' => 1,
		    'adj' => 1
		  );

my %cat_transfer = (
		    'auxEtre' => 'aux',
		    'auxAvoir' => 'aux'
		   );

sub normalize_cat {
  my $cat = shift;
  return $cat_transfer{$cat} || $cat;
}


my %value_transfer = (
		      'ssubj' => 'subj',
		      'vsubj' => 'subj'
		     );

sub normalize_value {
  my $v = shift;
  return $value_transfer{$v} || $v;
}

my %exclude_feature = (
##		       'lightverb' => 1,
		       'loc' => 1, # to be activated as soon as possible
		      );

my %feature_transfer = (
			'v-form' => 'mode',
			'form-aux' => 'form_aux',
			'aux-req' => 'aux_req',
			'demonstrative' => 'dem',
			'define' => 'def',
			'locative' => 'loc',
			'possessive' => 'poss',
			'inflnumber' => 'number',
			'inflperson' => 'person',
			'qu' => 'wh',
		       );

sub normalize_feature {
  my $f = shift;
  return $feature_transfer{$f} || $f;
}

my %pcas_transfer = (
		    );

sub normalize_pcas {
  my $p = shift;
  return $pcas_transfer{$p} || $p;
}

## use NumbersGram;
## my $nbparser = new NumbersGram();

#$::RD_TRACE=1;

#$::RD_AUTOACTION = q{ $item[1] };

#$::RD_HINT=1;

sub features {
  my @l = @_;
  my @fs;
  while (@l) {
    my $f = shift @l;
    my $v = shift @l;
    next if $exclude_feature{$f};
    push(@fs,$f => TAG::Comp->new( {name => $f,
				    rel => 'eq',
				    value => $v
				   }
				 ));
  }
  return @fs;
}

sub val {
  my $v = shift;
  return TAG::Val->new({text=>$v});
}

sub syntax {
  my $parser = shift;
  my $lemma = shift;
  my $subcat = shift;
  my $pos = 0;
  my $refl = ($lemma =~ s/Se$//o) ? TAG::Plus->new() : TAG::Minus->new() ;
  my $pos = 0;
  my %args = ();
  my @subcat = @$subcat;
#  print "SYNTAX $lemma $subcat\n";
#  print Data::Dumper->Dump([$subcat],['subcat']);
  foreach my $arg ('arg0' .. 'arg2') {
    $args{$arg} = subcat (shift @subcat || [ TAG::Minus->new() ]);
  }
  # print "SUBCAT ",dump(@args),"\n";
  my @args = features( lemma =>  [ val($lemma) ],
		       refl => [$refl],
		       %args
		     );
  my $ht = TAG::Feature->new({type=>'ht',f=> {@args }});
##  print "SUBCAT ",Data::Dumper->Dump([$ht],['ht']),"\n";
  return $ht;
}

sub subcat {
  my $vals = shift;
  my @pcas = ();
  my @vals = ();
  my %kind = ();
  if ($config->vcomp_expand()
      && grep( $_->{text} eq 'scomp', @$vals) 
      && !grep( $_->{text} eq 'vcomp', @$vals) ) {
    ## to overcome some current default of the lexicon
    push(@$vals,val('vcomp'),val('-vcomp'),val('de-vcomp'));
  }
  foreach my $val (@$vals) {
    next if ($val->{text} =~ /^arg\d/);
    if ($val->isa('TAG::Val')) {
      if  ($val->{text} =~ /^(\S+)-(\S+)/) {
	my $pcas = normalize_pcas($1);
	push(@pcas,val($pcas));
##	next if (@pcas > 1);	# no need to push another prepobj if already present
	$val = val("prep$2");
      }
    }
    next if (exists $kind{$val->{text}});
    $kind{$val->{text}} = 1;
    push(@vals,$val);
  }
  push(@pcas, TAG::Minus->new());
  return  [ 
	   TAG::Feature->new( { type => 'arg',
				f => { features( 'kind' => [@vals],
						 'pcas' => [@pcas]
					       )
				     }
			      } ) 
	  ];
}

sub entry {
  my $parser = shift;
  my $cat = shift;
  my $fs = shift;
  $fs->{type} = $cat;
  return TAG::LemmaRef->new({ cat => $cat, feature => $fs });
}

sub TAG::LemmaRef::lexLP {
  my $self = shift;
  my $name = TAG::LP::quote_aux($self->{'name'});
  my $cat = TAG::LP::quote_aux($self->{'cat'});
  my $lemma = '_';
  my $feature = $self->{'feature'};
  my $ht;
  if ($feature) {
    ## Special keys
    delete $feature->{f}{cat};

    my @ctrkeys = grep(/^(\S+-)?(ssubj|subj|obj|acomp|vcomp|scomp)/,keys %{$feature->{f}});

    foreach my $key (@ctrkeys) {
    }

    delete $feature->{f}{$_} foreach @ctrkeys;

    if(defined $feature->{f}{pred}) {
      $ht = $feature->{f}{pred}{value}->[0];
##      print Data::Dumper->Dump([$ht->{f}{lemma}{value}->[0]],['ht']);
      $lemma = TAG::LP::quote_aux($ht->{f}{lemma}{value}->[0]{'text'});
##      print "LEMMA $lemma\n";
      delete $ht->{f}{lemma};
      $ht = $ht->LP('ht');
      delete $feature->{f}{pred};
    }
    
    if (defined $feature->{f}{fct}) {
      my $subcat = $feature->{f}{fct}{value};
      $feature->{f}{fct} = TAG::Comp->new( {name => 'fct',
					    rel => 'eq',
					    value => subcat($subcat)
					    } );
    }

    # do not handle support verbs (lightverbs) at this time
    return  if (exists $feature->{f}{lightverb});

    $feature = $feature->LP($cat);
  } else {
    $feature = '_';
  }
  $ht = '_' unless defined $ht;
  return { lemma => $lemma, 
	   cat => $cat, 
	   top => $feature,
	   ht => $ht
	 }
}

sub entry_print {
    my ($self,$lex,$left,$right,$nlex) = @_;
    $lex = TAG::LP::quote_aux($lex);
    $nlex =~ s/qu'$/que/o;	# Should be properly handled by the lexicon !
    $nlex = TAG::LP::quote_aux($nlex);
    $nlex = $self->{lemma} unless ($self->{lemma} eq '_');
    print <<EOF
'C'($left, 
    lemma{ lex => $lex,
	   lemma => $nlex,
	   cat => $self->{cat},
	   top => $self->{top},
	   anchor  => tag_anchor{ name => $self->{ht} }
	 },
    $right
   ).

EOF
}

use LexedGram;
my $parser = new LexedGram();

read_macrofile($macrofile);

my @fifo = ();
my @pending = ();
my $nbsentence = 0;

my $lexin = '';
my $lexout = '';
my $h = start \@lexed,\$lexin,\$lexout;

while (<>) {
  segmenter($_);
}

$h->finish;

sub sentence_start {
  my $sentence = shift;
  @fifo = ();
  $nbsentence++;
  if ($expand) {
  print <<EOF;
%% Token database generated by $0
%% Sentence $nbsentence

EOF
} else {
  print "%% $sentence\n";
}
} 

sub sentence_end {
  my $n = shift;
  while (@fifo) {
    my $word = pop @fifo;
    handle_word($word) unless ($word->{sent});
  }
  print <<EOF if ($expand);

'N'($n).

EOF

  print <<EOF;
%% SENTENCE DIV %%

EOF

}

sub handle_date {
  my $date = TAG::LP::quote_aux(shift);
  my $lemma = TAG::LP::quote_aux(shift);
  my $left = shift;
  my $right = shift;
  print <<EOF if ($expand);
'C'($left,
    lemma{ lex => $date,
           cat => nc,
           top => nc{gender => masc, number => sg, time => (+)},
           lemma => $lemma
         },
    $right
   ).

EOF
}

sub handle_number {
  my $xnumber = shift;
  my $number = TAG::LP::quote_aux($xnumber);
  my $lemma = TAG::LP::quote_aux(shift);
  my $left = shift;
  my $right = shift;
  my $gender = shift || '_';
  send_word('_NUMBER',$xnumber,"pro|nc",$left,$right,'keep',undef);
  print <<EOF if ($expand);
'C'($left,
    lemma{ lex => $number,
           cat => number,
           top => number{ gender => $gender },
           lemma => $lemma
         },
    $right
   ).

EOF
}

sub handle_ordinal {
  my $lex = TAG::LP::quote_aux(shift);
  my $lemma = TAG::LP::quote_aux(shift);
  my $left = shift;
  my $right = shift;
  my $number = shift || '_';
  print <<EOF if ($expand);
'C'($left,
    lemma{ lex => $lex,
           cat => adj,
           top => adj{ number => $number },
           lemma => $lemma
         },
    $right
   ).

EOF
}


sub handle_word {
  my $data= shift;
  my $lex = $data->{lex};
  my $truelex = $data->{truelex};
  if (!exists $cachelexed{$lex}) {
    $lexin = "$lex\n";
    $h->pump until $lexout;
    my $info = $lexout;
    chomp $info;
    $cachelexed{$lex} = $info;
    $lexout = '';
  }
  my $info = $cachelexed{$lex};
  my $cat = $data->{cat};
  my %cat = ();
  if (defined $cat) {
    $cat{$_} ||= 1 foreach (split(/\|/,$cat));
  }
  my $left = $data->{left};
  my $right = $data->{right};
  my $status = $data->{status};
  my $altbox = $data->{altbox};
##  print "%% lexed handling '$lex' '$truelex' cat='$cat' status=$status: $info\n";
  if ($info =~ /<unknown>/ ) {
    return if ($status eq 'discard');
    return if (defined $altbox  && $altbox->{found});
    return unless ($expand);
    unless (defined $cat) {
      if ($lex =~ /^[[:upper:]]/ && $lex !~ /\b-[[:lower:]]/) {
	$cat = 'np';
      } else {
	$cat = 'cat[v,nc,adv,adj]';
	$unknown{$lex} = 1;
      }
    }
    $lex = TAG::LP::quote_aux($truelex);
    print <<EOF;
'C'($left, 
    lemma{ lex => $lex, cat => $cat },
    $right
   ).
EOF
    return;
  }
  return if (defined $altbox && $altbox->{found} == 1 && $data->{status} eq 'discard_but');
  if ($expand) {
    my $entries = entry_parse($info);
    foreach my $entry (@$entries) {
      next if (defined $cat && !exists $cat{$entry->{cat}});
      $altbox->{found} ||= 1 if (defined $altbox);
      entry_print($entry,$truelex,$left,$right,$lex);
    }
  } else {
    print "$data->{truelex}\n";
  }
}

sub entry_parse {
  my $entries = shift;
##  print "Entry parse: $entries\n";
  my $res  = $parser->entries($entries);
##  print Data::Dumper->Dump([$res],['entry']);
  return [ map($_->lexLP,@{$res}) ];
}

END {
  if (%unknown) {
    my $file = $config->tmpdir()."frmg.unknown";
    if (-r $file) {
      open(UNKNOWN,"<$file") || die "can't open $file: $!";
      foreach my $word (<UNKNOWN>) {
	chomp $word;
	delete $unknown{$word};
      }
      close(UNKNOWN);
    }
    open(UNKNOWN,">>$file") || die "can't open $file: $!";
    print UNKNOWN "$_\n" foreach (sort keys %unknown);
    close(UNKNOWN);
  }
}

sub segmenter {
  my $line = shift;
  my $i = 0;

  sentence_start($line);

  $line =~ s/([.]+)(?!:)/ $1/og;
  $line =~ s/([!?]+)(?!:)/ $1/og;
  $line =~ s/([\(\]+)/$1 /og;
  $line =~ s/([,;?\\)])(?!:)/ $1/og;
  $line =~ s/"/ " /og;
  $line =~ s/:\s+/ : /og;
##  $line =~ s/\bl'on\b/on/ogi;
  $line =~ s/((?:(?:^|\s)\w|qu)\')/$1 /ogi;
  $line =~ s/\b-t-(il|elle|on|y|ils|elles)\b/ -t-$1/ogi;
  $line =~  s/\B([td])-(il|elle|on|y|ils|elles)\b/$1 -$2/ogi;
  $line =~ s/\B-je\b/e -je/ogi;
  $line =~ s/(?<!-t)-(je|tu|nous|vous|ce|ci|mme|le|la|lui|moi|toi|leur|on|en)
	     (?=$|\s+|-(?:le|la|lui|moi|toi|leur|on|en))
	     / -$1/ogix;
  $line =~ s/\b-(ci|l)\b/ -$1/ogi;

 ## $line =~ s/\b($agglpat)\b/$agglutinates{lc($1)}/oig;
 ##  $line =~ s/\b($agglpat)-/$agglutinates{lc($1)}-/oig;

  ##      print STDERR $line;
  ##      $line =~ //g;
  ## print "%% LINE '$line'\n";
  
 LOOP: {
    
##    print "%% HERE $i\n";

    redo LOOP if ($line =~ /\G\s+/gc);
    
    last LOOP if ($line =~ /\G\#/gc || $line =~ /\G$/gc);
    my $left = $i;
    my $right = ++$i;

    ## Roman Number
    if ($line =~ /\G([IVXLCDM]{2,}(:?e|eme(:?s)))(\s+|$)/gc) {
      my $ordinal = $1;
      my $number = ($ordinal =~ /s$/) ? 'pl' : 'sg';
      handle_ordinal($ordinal,'ROMAN_NUMBER',$left,$right,$number);
      redo LOOP;
    }
   
    ## Dates
    if ($line =~ /\G($RE->{date})\b/gc) {
      my $date = $1;
      $i = $right += split(/\s+/,$date) - 1;
      handle_date($1,'DATE',$left,$right);
      redo LOOP;
    }
    
    ## Ordinals
    if ($line =~ /\G($RE->{ordinal})\b/gc) {
      my $ordinal = $1;
      $i = $right += split(/\s+/,$ordinal) - 1;
      my $number = ($ordinal =~ /s$/) ? 'pl' : 'sg';
      handle_ordinal($ordinal,'ORDINAL',$left,$right,$number);
      redo LOOP;
    }

    ## Other numbers
    if ($line =~ /\G($RE->{number})\b/gc) {
      my $number = $1;
      $i = $right += split(/\s+/,$number) - 1;
      my $gender = ($number =~ /une$/) ? 'fem' : '_';
      handle_number($number,'NUMBER',$left,$right,$gender);
      redo LOOP;
    }

    if ($line =~ /\G:\s+/gc) {
      send_word(':',':',undef,$left,$right,'keep',undef);
        redo LOOP;
    }

    ## agglutinates
    if ($line =~ /\G($agglpat)\b/gc) {
        my $word = $1;
        my @words = split(/\s+/,$agglutinates{lc($word)});
        $i = $right += @words - 1;
	my $altbox ||= { found => 0 };
	cut_word($word,$left,$right,$altbox,'discard');
	for (my $j = $left; @words ; $j++) {
	  my $word = shift @words;
	  cut_word($word,$j,$j+1,$altbox);
	}
        redo LOOP;
    }

    ## standard tokens
    if ($line =~ /\G(\S+)/gc){
      cut_word($1,$left,$right);
      redo LOOP;
    }
  }

  sentence_end($i);

}

sub cut_word {
  my ($xlex,$left,$right,$altbox,$status) = @_;
  my ($lex,$cat,$ht,$top) = split(':',$xlex);
  my $truelex = $lex;
  $lex =~ s/(?<!_)_(?!_)/ /o;
  unless ($status) {
    $status = 'keep';
    $status = 'discard' if ($lex =~ m{/\w+/});
  }
  if ($left == 0 && $lex =~ /^[[:upper:]]/) {
    my $llex = lc($lex); 
    $altbox ||= { found => 0 };
    if ($llex eq 'a') {
      send_word('',$truelex,$cat,$left,$right,'discard',$altbox);
    }
    ##        my $lcat = $cat || "det|pri|prep|adv";
    send_word($llex,$truelex,$cat,$left,$right,'discard',$altbox);
    send_word($lex,$truelex,$cat,$left,$right,'keep',$altbox);
  } else {
    send_word($lex,$truelex,$cat,$left,$right,$status,$altbox);
  }
}

sub send_word {
  my ($lex,$truelex,$cat,$left,$right,$status,$altbox) = @_;
  ##  print "%% preparing sending $lex: @fifo\n";
  my @first = ($#fifo > 5 ? @fifo[0 .. 4] : @fifo);
  foreach my $word (@first) {
    ##    print "%% trying merge\n";
    if ($word->{right} == $left) {
      my $newlex = "$word->{lex} $lex";
      $newlex =~ s/' /'/og;
      if (exists $locutions{$newlex} || exists $locutions{lc($newlex)}) {
	send_word($newlex,$newlex,undef,$word->{left},$right,'discard',$altbox);
      } elsif (!$word->{sent}) {
	$word->{sent} = 1;
	handle_word($word);
      }
    }
  }
##  print "%% sending $lex ($truelex)\n";
  my $locution = (exists $locutions{$lex}) || (exists $locutions{lc($lex)});
  my $data = {
	      lex => $lex,
	      truelex => $truelex,
	      cat => $cat,
	      left => $left,
	      right => $right,
	      status => $status,
	      altbox => $altbox,
	      sent => 0
	     };
  if ($locution) {
    ## we wait until next word before deciding trying $lex
    $data->{altbox} = $altbox ||= {found => 0 };
##    $data->{status} = 'discard';
    unshift(@fifo,$data);
    return;
  }
  if ($lex =~ /^(?:$prefpat)([^-]+)/) {
    my $seed = $1;
##    print "Try compound $lex seed=$seed\n";
    $data->{altbox} = $altbox ||= {found => 0};
    $data->{status} = 'discard';
    handle_word($data);
    my $seeddata = {
		    lex => $seed,
		    truelex => $truelex,
		    cat => $cat,
		    left => $left,
		    right => $right,
		    status => 'discard_but',
		    altbox => $altbox,
		    sent => 0
		   };
    handle_word($seeddata);
  } else {
    handle_word($data);
  }
}

sub read_macrofile {
  my $macrofile = shift;
  open(MACRO,"<$macrofile") || die "can't open $macrofile";
  while (<MACRO>) {
    next unless /^\@/;
    my $line = $_;
    $line =~ s{(/{2}.*)$}{};
    chomp $line;
    my ($name,$body) = split(/\s*=\s*/,$line,2);
    next unless ($body =~ /^\[/);
    ##    print "PARSING MACRO $name $body\n";
    ##    my $fs = $parser->macrodef(\$body);
    ##    print dump($fs->{f});
    ##    $macro{$name} = [ %{$fs->{f}} ];
    $macro{$name} = $body;
  }
  close(MACRO);
  return %macro;
}

sub macro_expand ($) {
  my $macro = shift;
  return [] unless exists $macro{$macro};
  my $fs = $parser->macrodef($macro{$macro});
##  print "expand macro $macro: $macro{$macro} -> $fs\n";
  return [ %{$fs->{f}} ];
}

sub quote {
  my $string = shift;
  return $string if ($string =~ /^[a-z_]\w*$/o);
  return $string if ($string =~ /^\d+$/);
  #    $string =~ s/\'/\'\'/og;
  $string =~ s/\'(?!\')/\'\'/og;
  return "\'$string\'";
}

sub parse_map_file {
    my $map = shift;
    my $file = shift;
    open("MAP","<$file") || die "can't open map file $file";
    while (<MAP>) {
        chop;
        next if (/^\#/);
        next if (/^\s*$/);
        $map->{$1}{$2} = $3, next if (/^(\w+)\s+(\S+)\s*:\s*(.+)\s*$/);
        $map->{$1} = $2, next if (/^(\w+)\s*:\s*(.+)\s*$/);
    }
    close(MAP);
}


1;

__END__
